# Copyright AGNTCY Contributors (https://github.com/agntcy)
# SPDX-License-Identifier: Apache-2.0

import asyncio
import datetime
import logging
import sys
import time
from collections.abc import AsyncGenerator, AsyncIterable, Callable
from typing import Any

if sys.version_info >= (3, 11):
    from asyncio import timeout_at as asyncio_timeout_at
else:
    from async_timeout import timeout_at as asyncio_timeout_at

import slim_bindings
from google.rpc import code_pb2, status_pb2

from slimrpc.common import (
    DEADLINE_KEY,
    MAX_TIMEOUT,
    RequestType,
    ResponseType,
    SLIMAppConfig,
    create_local_app,
    service_and_method_to_pyname,
    split_id,
)
from slimrpc.rpc import SRPCResponseError

logger = logging.getLogger(__name__)


class Channel:
    def __init__(
        self,
        remote: str,
        local_app: slim_bindings.Slim,
    ) -> None:
        self.remote = split_id(remote)
        self.local_app = local_app

    @classmethod
    async def from_slim_app_config(
        cls, remote: str, slim_app_config: SLIMAppConfig
    ) -> "Channel":
        local_app = await create_local_app(slim_app_config)
        return cls(remote=remote, local_app=local_app)

    async def _common_setup(
        self, method: str, metadata: dict[str, str] | None = None
    ) -> tuple[slim_bindings.Name, slim_bindings.Session, dict[str, str]]:
        service_name = service_and_method_to_pyname(self.remote, method)

        await self.local_app.set_route(
            service_name,
        )

        logger.info(f"creating session for service {service_name}")

        # Create a session using PointToPoint configuration
        session, ack = await self.local_app.create_session(
            destination=service_name,
            session_config=slim_bindings.SessionConfiguration.PointToPoint(
                max_retries=10,
                timeout=datetime.timedelta(seconds=1),
            ),
        )
        await ack

        return service_name, session, metadata or {}

    async def _delete_session(self, session: slim_bindings.Session) -> None:
        try:
            await self.local_app.delete_session(session)
        except Exception as e:
            logger.warn(f"Error deleting session: {e}")

    async def _send_unary(
        self,
        request: RequestType,
        session: slim_bindings.Session,
        service_name: slim_bindings.Name,
        metadata: dict[str, str],
        request_serializer: Callable,
        deadline: float,
    ) -> None:
        # Add deadline to metadata
        metadata[DEADLINE_KEY] = str(deadline)

        # Send the request
        request_bytes = request_serializer(request)
        ack = await session.publish(
            request_bytes,
            metadata=metadata,
        )
        await ack

    async def _send_stream(
        self,
        request_stream: AsyncIterable,
        session: slim_bindings.Session,
        service_name: slim_bindings.Name,
        metadata: dict[str, str],
        request_serializer: Callable,
        deadline: float,
    ) -> None:
        # Add deadline to metadata
        metadata[DEADLINE_KEY] = str(deadline)

        # Send requests
        async for request in request_stream:
            request_bytes = request_serializer(request)
            ack = await session.publish(
                request_bytes,
                metadata=metadata,
            )
            await ack

        # Send end of streaming message
        ack = await session.publish(
            b"",
            metadata={**metadata, "code": str(code_pb2.OK)},
        )
        await ack

    async def _receive_unary(
        self,
        session: slim_bindings.Session,
        response_deserializer: Callable,
        deadline: float,
    ) -> tuple[slim_bindings.MessageContext, Any]:
        # Wait for the response
        async with asyncio_timeout_at(
            _compute_loop_deadline_from_real_deadline(deadline)
        ):
            msg_ctx, response_bytes = await session.get_message()

            code = msg_ctx.metadata.get("code")
            if code != str(code_pb2.OK):
                status = status_pb2.Status.FromString(response_bytes)
                raise SRPCResponseError(status.code, status.message, status.details)

            response = response_deserializer(response_bytes)
            return msg_ctx, response

    async def _receive_stream(
        self,
        session: slim_bindings.Session,
        response_deserializer: Callable,
        deadline: float,
    ) -> AsyncIterable:
        # Wait for the responses
        async def generator() -> AsyncIterable:
            try:
                while True:
                    msg_ctx, response_bytes = await session.get_message()

                    code = msg_ctx.metadata.get("code")
                    if code != str(code_pb2.OK):
                        status = status_pb2.Status.FromString(response_bytes)
                        raise SRPCResponseError(
                            status.code, status.message, status.details
                        )

                    if not response_bytes:
                        logger.debug("end of stream received")
                        break

                    response = response_deserializer(response_bytes)
                    yield response
            except SRPCResponseError:
                raise
            except Exception as e:
                logger.error(f"error receiving messages: {e}")
                raise

        async with asyncio_timeout_at(
            _compute_loop_deadline_from_real_deadline(deadline)
        ):
            async for response in generator():
                yield response

    def stream_stream(
        self,
        method: str,
        request_serializer: Callable = lambda x: x,
        response_deserializer: Callable = lambda x: x,
    ) -> Callable:
        async def call_stream_stream(
            request_stream: AsyncIterable,
            timeout: int = MAX_TIMEOUT,
            metadata: dict | None = None,
        ) -> AsyncIterable:
            try:
                service_name, session, metadata = await self._common_setup(
                    method, metadata
                )

                deadline = _compute_real_deadline(timeout)

                # Send the requests
                await self._send_stream(
                    request_stream,
                    session,
                    service_name,
                    metadata,
                    request_serializer,
                    deadline,
                )

                # Wait for the responses
                async for response in self._receive_stream(
                    session,
                    response_deserializer,
                    deadline,
                ):
                    yield response
            finally:
                await self._delete_session(session)

        return call_stream_stream

    def stream_unary(
        self,
        method: str,
        request_serializer: Callable = lambda x: x,
        response_deserializer: Callable = lambda x: x,
    ) -> Callable:
        async def call_stream_unary(
            request_stream: AsyncIterable,
            timeout: int = MAX_TIMEOUT,
            metadata: dict | None = None,
        ) -> ResponseType:
            try:
                service_name, session, metadata = await self._common_setup(
                    method, metadata
                )

                deadline = _compute_real_deadline(timeout)

                # Send the requests
                await self._send_stream(
                    request_stream,
                    session,
                    service_name,
                    metadata,
                    request_serializer,
                    deadline,
                )

                # Wait for response
                _, ret = await self._receive_unary(
                    session,
                    response_deserializer,
                    deadline,
                )

                return ret
            finally:
                await self._delete_session(session)

        return call_stream_unary

    def unary_stream(
        self,
        method: str,
        request_serializer: Callable = lambda x: x,
        response_deserializer: Callable = lambda x: x,
    ) -> Callable:
        async def call_unary_stream(
            request: RequestType,
            timeout: int = MAX_TIMEOUT,
            metadata: dict[str, str] | None = None,
        ) -> AsyncGenerator:
            try:
                service_name, session, metadata = await self._common_setup(
                    method, metadata
                )

                deadline = _compute_real_deadline(timeout)

                # Send the request
                await self._send_unary(
                    request,
                    session,
                    service_name,
                    metadata,
                    request_serializer,
                    deadline,
                )

                # Wait for the responses
                async for response in self._receive_stream(
                    session,
                    response_deserializer,
                    deadline,
                ):
                    yield response
            finally:
                await self._delete_session(session)

        return call_unary_stream

    def unary_unary(
        self,
        method: str,
        request_serializer: Callable = lambda x: x,
        response_deserializer: Callable = lambda x: x,
    ) -> Callable:
        async def call_unary_unary(
            request: RequestType,
            timeout: int = MAX_TIMEOUT,
            metadata: dict[str, str] | None = None,
        ) -> ResponseType:
            try:
                service_name, session, metadata = await self._common_setup(
                    method, metadata
                )

                deadline = _compute_real_deadline(timeout)

                # Send request
                await self._send_unary(
                    request,
                    session,
                    service_name,
                    metadata,
                    request_serializer,
                    deadline,
                )

                # Wait for the response
                _, ret = await self._receive_unary(
                    session,
                    response_deserializer,
                    deadline,
                )

                return ret
            finally:
                await self._delete_session(session)

        return call_unary_unary


def _compute_real_deadline(timeout: int) -> float:
    return time.time() + float(timeout)


def _compute_loop_deadline_from_real_deadline(real_deadline: float) -> float:
    return asyncio.get_running_loop().time() + (real_deadline - time.time())



# Copyright AGNTCY Contributors (https://github.com/agntcy)
# SPDX-License-Identifier: Apache-2.0

import base64
import json
import logging
from dataclasses import dataclass
from typing import Tuple, TypeVar

import slim_bindings

logger = logging.getLogger(__name__)

DEADLINE_KEY = "slimrpc-timeout"

MAX_TIMEOUT = 36000  # 10h

# Types for SRPC API
RequestType = TypeVar("RequestType")
ResponseType = TypeVar("ResponseType")


class color:
    PURPLE = "\033[95m"
    CYAN = "\033[96m"
    DARKCYAN = "\033[36m"
    BLUE = "\033[94m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    END = "\033[0m"


# Split an ID into its components
# Expected format: organization/namespace/application
# Raises ValueError if the format is incorrect
# Returns a Name with the 3 components
def split_id(id: str) -> slim_bindings.Name:
    try:
        organization, namespace, app = id.split("/")
    except ValueError as e:
        print("Error: IDs must be in the format organization/namespace/app-or-stream.")
        raise e

    return slim_bindings.Name(organization, namespace, app)


def method_to_pyname(
    name: slim_bindings.Name, service_name: str, method_name: str
) -> slim_bindings.Name:
    """
    Convert a method name to a Name.
    """

    components = name.components_strings()

    if len(components) < 3:
        raise ValueError("Name must have at least 3 components.")

    subscription_name = slim_bindings.Name(
        components[0],
        components[1],
        f"{components[2]}-{service_name}-{method_name}",
    )

    logger.debug(f"Name after conversion from service/method: {subscription_name}")

    return subscription_name


def service_and_method_to_pyname(
    name: slim_bindings.Name, service_method: str
) -> slim_bindings.Name:
    """
    Convert a method name to a Name.
    """

    # Split method in service and method name
    service_name = service_method.split("/")[1]
    method_name = service_method.split("/")[2]

    return method_to_pyname(name, service_name, method_name)


def handler_name_to_pyname(
    name: slim_bindings.Name,
    service_name: str,
    method_name: str,
) -> slim_bindings.Name:
    """
    Convert a handler name to a Name.
    """

    return method_to_pyname(name, service_name, method_name)


# Create a shared secret identity provider and verifier
# This is used for shared secret authentication
# Takes an identity and a shared secret as parameters
# Returns a tuple of (provider, verifier)
# This is used for shared secret authentication
def shared_secret_identity(
    identity: str, secret: str
) -> Tuple[slim_bindings.IdentityProvider, slim_bindings.IdentityVerifier]:
    """
    Create a provider and verifier using a shared secret.
    """
    provider = slim_bindings.IdentityProvider.SharedSecret(
        identity=identity, shared_secret=secret
    )
    verifier = slim_bindings.IdentityVerifier.SharedSecret(
        identity=identity, shared_secret=secret
    )

    return provider, verifier


# Create a JWT identity provider and verifier
# This is used for JWT authentication
# Takes private key path, public key path, and algorithm as parameters
# Returns a Slim object with the provider and verifier
def jwt_identity(
    jwt_path: str,
    jwk_path: str,
    iss: str | None = None,
    sub: str | None = None,
    aud: list | None = None,
) -> Tuple[slim_bindings.IdentityProvider, slim_bindings.IdentityVerifier]:
    """
    Parse the JWK and JWT from the provided strings.
    """

    print(f"Using JWk file: {jwk_path}")

    with open(jwk_path) as jwk_file:
        jwk_string = jwk_file.read()

    # The JWK is normally encoded as base64, so we need to decode it
    spire_jwks = json.loads(jwk_string)

    for _, v in spire_jwks.items():
        # Decode first item from base64
        spire_jwks = base64.b64decode(v)
        break

    provider = slim_bindings.IdentityProvider.StaticJwt(
        path=jwt_path,
    )

    pykey = slim_bindings.Key(
        algorithm=slim_bindings.Algorithm.RS256,
        format=slim_bindings.KeyFormat.Jwks,
        key=slim_bindings.KeyData.Content(content=spire_jwks.decode("utf-8")),
    )

    verifier = slim_bindings.IdentityVerifier.Jwt(
        public_key=pykey,
        issuer=iss,
        audience=aud,
        subject=sub,
    )

    return provider, verifier


@dataclass
class SLIMAppConfig:
    identity: str
    slim_client_config: dict
    enable_opentelemetry: bool = False
    shared_secret: str = ""

    def identity_pyname(self) -> slim_bindings.Name:
        return split_id(self.identity)


async def create_local_app(config: SLIMAppConfig) -> slim_bindings.Slim:
    # init tracing
    slim_bindings.init_tracing(
        {
            "log_level": "info",
            "opentelemetry": {
                "enabled": config.enable_opentelemetry,
                "grpc": {
                    "endpoint": "http://localhost:4317",
                },
            },
        }
    )

    split_identity = config.identity_pyname()

    provider, verifier = shared_secret_identity(
        identity=str(split_identity),
        secret=config.shared_secret,
    )

    local_app = slim_bindings.Slim(split_identity, provider, verifier)

    logger.info(f"{local_app.local_name} Created app")

    # Connect to slim server
    _ = await local_app.connect(config.slim_client_config)

    logger.info(
        f"{local_app.local_name} Connected to {config.slim_client_config['endpoint']}"
    )

    return local_app



import asyncio
import logging
from collections.abc import AsyncIterable

from slimrpc.common import SLIMAppConfig
from slimrpc.context import MessageContext, SessionContext
from slimrpc.examples.simple.types.example_pb2 import ExampleRequest, ExampleResponse
from slimrpc.examples.simple.types.example_pb2_slimrpc import (
    TestServicer,
    add_TestServicer_to_server,
)
from slimrpc.server import Server

logger = logging.getLogger(__name__)


class TestService(TestServicer):
    async def ExampleUnaryUnary(
        self,
        request: ExampleRequest,
        msg_context: MessageContext,
        session_context: SessionContext,
    ) -> ExampleResponse:
        logger.info(f"Received unary-unary request: {request}")

        return ExampleResponse(example_integer=1, example_string="Hello, World!")

    async def ExampleUnaryStream(
        self,
        request: ExampleRequest,
        msg_context: MessageContext,
        session_context: SessionContext,
    ) -> AsyncIterable[ExampleResponse]:
        logger.info(f"Received unary-stream request: {request}")

        # generate async responses stream
        for i in range(5):
            logger.info(f"Sending response {i}")
            yield ExampleResponse(example_integer=i, example_string=f"Response {i}")

    async def ExampleStreamUnary(
        self,
        request_iterator: AsyncIterable[tuple[ExampleRequest, MessageContext]],
        session_context: SessionContext,
    ) -> ExampleResponse:
        logger.info(f"Received stream-unary request: {request_iterator}")

        received_strs = []
        async for request, msg_ctx in request_iterator:
            _ = msg_ctx  # Unused in this example
            received_strs.append(request.example_string)

        response = ExampleResponse(
            example_integer=len(received_strs),
            example_string="Saw: " + ", ".join(received_strs),
        )
        return response

    async def ExampleStreamStream(
        self,
        request_iterator: AsyncIterable[tuple[ExampleRequest, MessageContext]],
        session_context: MessageContext,
    ) -> AsyncIterable[ExampleResponse]:
        logger.info(f"Received stream-stream request: {request_iterator}")

        async for request, msg_ctx in request_iterator:
            _ = msg_ctx  # Unused in this example
            yield ExampleResponse(
                example_integer=request.example_integer * 100,
                example_string=f"Echo: {request.example_string}",
            )


async def amain() -> None:
    server = await Server.from_slim_app_config(
        slim_app_config=SLIMAppConfig(
            identity="agntcy/grpc/server",
            slim_client_config={
                "endpoint": "http://localhost:46357",
                "tls": {"insecure": True},
            },
            enable_opentelemetry=False,
            shared_secret="my_shared_secret_for_testing_purposes_only",
        )
    )

    # Create RPCs
    add_TestServicer_to_server(
        TestService(),
        server,
    )

    await server.run()


def main() -> None:
    """
    Main entry point for the server.
    """
    logging.basicConfig(level=logging.DEBUG)
    try:
        asyncio.run(amain())
    except KeyboardInterrupt:
        print("Server interrupted by user.")



import asyncio
import logging
from collections.abc import AsyncGenerator

import slimrpc
from slimrpc.examples.simple.types.example_pb2 import ExampleRequest
from slimrpc.examples.simple.types.example_pb2_slimrpc import TestStub

logger = logging.getLogger(__name__)


async def amain() -> None:
    channel = await slimrpc.Channel.from_slim_app_config(
        remote="agntcy/grpc/server",
        slim_app_config=slimrpc.SLIMAppConfig(
            identity="agntcy/grpc/client",
            slim_client_config={
                "endpoint": "http://localhost:46357",
                "tls": {
                    "insecure": True,
                },
            },
            enable_opentelemetry=False,
            shared_secret="my_shared_secret_for_testing_purposes_only",
        ),
    )

    # Stubs
    stubs = TestStub(channel)

    # Call method
    try:
        request = ExampleRequest(example_integer=1, example_string="hello")
        response = await stubs.ExampleUnaryUnary(request, timeout=2)

        logger.info(f"Response: {response}")

        responses = stubs.ExampleUnaryStream(request, timeout=2)
        async for resp in responses:
            logger.info(f"Stream Response: {resp}")

        async def stream_requests() -> AsyncGenerator[ExampleRequest, None]:
            for i in range(10):
                yield ExampleRequest(example_integer=i, example_string=f"Request {i}")

        response = await stubs.ExampleStreamUnary(stream_requests(), timeout=2)
        logger.info(f"Stream Unary Response: {response}")

        stream = stubs.ExampleStreamStream(stream_requests(), timeout=2)
        async for resp in stream:
            logger.info(f"Stream Stream Response: {resp}")
        logger.info("Stream Stream completed")

    except asyncio.TimeoutError:
        logger.exception("timeout while waiting for response")

    await asyncio.sleep(1)


def main() -> None:
    """
    Main entry point for the server.
    """
    logging.basicConfig(level=logging.INFO)
    try:
        asyncio.run(amain())
    except KeyboardInterrupt:
        print("Server interrupted by user.")



# Copyright AGNTCY Contributors (https://github.com/agntcy)
# SPDX-License-Identifier: Apache-2.0

from google.rpc.code_pb2 import Code as StatusCode

from slimrpc.channel import Channel, SLIMAppConfig
from slimrpc.context import Context
from slimrpc.rpc import (
    RPCHandler,
    SRPCResponseError,
    stream_stream_rpc_method_handler,
    stream_unary_rpc_method_handler,
    unary_stream_rpc_method_handler,
    unary_unary_rpc_method_handler,
)
from slimrpc.server import Server

__all__ = [
    "StatusCode",
    "Context",
    "SRPCResponseError",
    "RPCHandler",
    "stream_stream_rpc_method_handler",
    "stream_unary_rpc_method_handler",
    "unary_stream_rpc_method_handler",
    "unary_unary_rpc_method_handler",
    "Server",
    "Channel",
    "SLIMAppConfig",
]



# Copyright AGNTCY Contributors (https://github.com/agntcy)
# SPDX-License-Identifier: Apache-2.0

import asyncio
import logging
import sys
import time
from collections.abc import AsyncGenerator, Callable
from dataclasses import dataclass
from typing import Any, Tuple

if sys.version_info >= (3, 11):
    from asyncio import timeout as asyncio_timeout
else:
    from async_timeout import timeout as asyncio_timeout

import slim_bindings
from google.rpc import code_pb2, status_pb2

from slimrpc.common import (
    DEADLINE_KEY,
    MAX_TIMEOUT,
    SLIMAppConfig,
    create_local_app,
    handler_name_to_pyname,
)
from slimrpc.context import MessageContext, SessionContext
from slimrpc.rpc import RPCHandler, SRPCResponseError

logger = logging.getLogger(__name__)


@dataclass(unsafe_hash=True)
class ServiceMethod:
    service: str
    method: str


class Server:
    def __init__(
        self,
        local_app: slim_bindings.Slim,
    ) -> None:
        self._local_app = local_app
        self.handlers: dict[ServiceMethod, RPCHandler] = {}
        self._pyname_to_handler: dict[slim_bindings.Name, RPCHandler] = {}

    @classmethod
    async def from_slim_app_config(cls, slim_app_config: SLIMAppConfig) -> "Server":
        local_app = await create_local_app(slim_app_config)
        return cls(local_app=local_app)

    def register_method_handlers(
        self, service_name: str, handlers: dict[str, RPCHandler]
    ) -> None:
        """
        Register method handlers for the server.
        """
        for method_name, handler in handlers.items():
            self.register_rpc(service_name, method_name, handler)

    def register_rpc(
        self, service_name: str, method_name: str, rpc_handler: RPCHandler
    ) -> None:
        """
        Register an RPC handler for the server.
        """

        # Register the RPC handler
        self.handlers[ServiceMethod(service=service_name, method=method_name)] = (
            rpc_handler
        )

    async def run(self) -> None:
        """
        Run the server, creating a local SLIM instance and subscribing to the handlers.
        """

        logger.info(f"Subscribing to {self._local_app.local_name.id}")

        await self._local_app.subscribe(self._local_app.local_name)
        logger.info(
            f"Subscribing to {self._local_app.local_name}",
        )

        # Subscribe and store mapping to subscription name for session handling
        # later.
        self._pyname_to_handler = {}
        for service_method, handler in self.handlers.items():
            subscription_name = handler_name_to_pyname(
                self._local_app.local_name,
                service_method.service,
                service_method.method,
            )
            strs = subscription_name.components_strings()
            s_clone = slim_bindings.Name(
                strs[0], strs[1], strs[2], self._local_app.local_name.id
            )
            logger.info(
                f"Subscribing to {s_clone}",
            )
            await self._local_app.subscribe(s_clone)
            self._pyname_to_handler[s_clone] = handler

        instance = self._local_app.id_str

        # Wait for a message and reply in a loop
        while True:
            logger.info(f"{instance} waiting for new session to be established")

            session = await self._local_app.listen_for_session()
            logger.info(
                f"{instance} received a new session: {session.id}",
            )

            asyncio.create_task(self.handle_session(session))

    async def handle_session(
        self,
        session: slim_bindings.Session,
    ) -> None:
        logger.info(f"new session from {session.dst} to {session.src}")

        # Call the RPC handler
        try:
            # Find the RPC handler for the session
            rpc_handler: RPCHandler | None = self._pyname_to_handler.get(
                session.src, None
            )
            if rpc_handler is None:
                logger.error(
                    f"no handler found for session {session.id} with destination {session.src}",
                )
                return

            # Get deadline from request
            deadline_str = session.metadata.get(DEADLINE_KEY, "")
            timeout = (
                max(float(deadline_str) - time.time(), 0)
                if deadline_str
                else MAX_TIMEOUT
            )

            async with asyncio_timeout(timeout):
                # Send the response back to the client
                async for code, response in call_handler(
                    rpc_handler,
                    session,
                ):
                    logger.info(
                        f"sending response for session {session.id} with code {code}"
                    )
                    ack = await session.publish(
                        rpc_handler.response_serializer(response),
                        metadata={"code": str(code)},
                    )
                    await ack

                # Send end of stream message if the response was streaming
                if rpc_handler.response_streaming:
                    ack = await session.publish(
                        b"",
                        metadata={"code": str(code_pb2.OK)},
                    )
                    await ack
        except Exception as e:
            if isinstance(e, asyncio.TimeoutError):
                logger.warn(f"session {session.id} timed out after {timeout} seconds")
            else:
                logger.error(f"error handling session {session.id}: {e}")

            # Only close the session if there is an uncaught exception, most
            # errors should be caught and bubbled up to the client in the RPC
            # communication, but if there is another issue or we timeout we
            # close our own session because it'll let the client know something
            # went wrong.
            await self.try_close_session(session)

    async def try_close_session(self, session: slim_bindings.Session) -> None:
        try:
            await self._local_app.delete_session(session)
        except Exception as e:
            logger.error(f"Error closing session: {e}")


async def request_generator(
    session: slim_bindings.Session,
    request_deserializer: Callable,
) -> AsyncGenerator[Any, MessageContext]:
    try:
        while True:
            msg_ctx, request_bytes = await session.get_message()

            if msg_ctx.metadata.get("code") == str(code_pb2.OK) and not request_bytes:
                logger.debug("end of stream received")
                break

            request = request_deserializer(request_bytes)
            context = MessageContext.from_message_context(msg_ctx)
            yield request, context
    except Exception as e:
        logger.error(f"Error receiving messages from SLIM: {e}")
        raise


async def call_handler(
    handler: RPCHandler,
    session: slim_bindings.Session,
) -> AsyncGenerator[Tuple[Any, Any], None]:
    """
    Call the handler with the given arguments.
    """
    try:
        session_context = SessionContext.from_session(session)

        handler_args: Any
        if not handler.request_streaming:
            # For unary requests, get the single message
            msg_ctx, request_bytes = await session.get_message()
            request = handler.request_deserializer(request_bytes)
            handler_args = (
                request,
                MessageContext.from_message_context(msg_ctx),
                session_context,
            )
        else:
            # For streaming requests, create a generator
            request_iterator = request_generator(session, handler.request_deserializer)
            handler_args = (request_iterator, session_context)

        if not handler.response_streaming:
            # If not streaming we expect a single response and then exit the generator.
            response = await handler.behaviour(*handler_args)
            yield code_pb2.OK, response
            return

        # For streaming responses yield each response from the handler
        async for response in handler.behaviour(*handler_args):
            yield code_pb2.OK, response
    except SRPCResponseError as e:
        yield (
            e.code,
            status_pb2.Status(code=e.code, message=e.message, details=e.details),
        )
    except Exception:
        logger.exception("Unexpected error while calling handler")
        yield (
            code_pb2.UNKNOWN,
            status_pb2.Status(
                code=code_pb2.UNKNOWN, message="Internal Error", details=None
            ),
        )



# Generated by protoc-slimrpc-plugin
# This file is generated from an embedded template. Do not edit directly.

import slimrpc
from slimrpc import rpc as slimrpc_rpc

from google.rpc import code_pb2 as code__pb2

from . import example_pb2 as pb2


class TestStub:
    """Client stub for Test."""
    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A slimrpc.Channel.
        """
        self.ExampleUnaryUnary = channel.unary_unary(
            "/example_service.Test/ExampleUnaryUnary",
            request_serializer=pb2.ExampleRequest.SerializeToString,
            response_deserializer=pb2.ExampleResponse.FromString,
        )
        self.ExampleUnaryStream = channel.unary_stream(
            "/example_service.Test/ExampleUnaryStream",
            request_serializer=pb2.ExampleRequest.SerializeToString,
            response_deserializer=pb2.ExampleResponse.FromString,
        )
        self.ExampleStreamUnary = channel.stream_unary(
            "/example_service.Test/ExampleStreamUnary",
            request_serializer=pb2.ExampleRequest.SerializeToString,
            response_deserializer=pb2.ExampleResponse.FromString,
        )
        self.ExampleStreamStream = channel.stream_stream(
            "/example_service.Test/ExampleStreamStream",
            request_serializer=pb2.ExampleRequest.SerializeToString,
            response_deserializer=pb2.ExampleResponse.FromString,
        )


class TestServicer():
    """Server servicer for Test. Implement this class to provide your service logic."""

    def ExampleUnaryUnary(self, request, msg_context, session_context):
        """Method for ExampleUnaryUnary. Implement your service logic here."""
        raise slimrpc_rpc.SRPCResponseError(
            code=code__pb2.UNIMPLEMENTED, message="Method not implemented!"
        )
    def ExampleUnaryStream(self, request, msg_context, session_context):
        """Method for ExampleUnaryStream. Implement your service logic here."""
        raise slimrpc_rpc.SRPCResponseError(
            code=code__pb2.UNIMPLEMENTED, message="Method not implemented!"
        )
    def ExampleStreamUnary(self, request_iterator, session_context):
        """Method for ExampleStreamUnary. Implement your service logic here."""
        raise slimrpc_rpc.SRPCResponseError(
            code=code__pb2.UNIMPLEMENTED, message="Method not implemented!"
        )
    def ExampleStreamStream(self, request_iterator, session_context):
        """Method for ExampleStreamStream. Implement your service logic here."""
        raise slimrpc_rpc.SRPCResponseError(
            code=code__pb2.UNIMPLEMENTED, message="Method not implemented!"
        )




def add_TestServicer_to_server(servicer, server: slimrpc.Server):
    rpc_method_handlers = {
        "ExampleUnaryUnary": slimrpc.unary_unary_rpc_method_handler(
            behaviour=servicer.ExampleUnaryUnary,
            request_deserializer=pb2.ExampleRequest.FromString,
            response_serializer=pb2.ExampleResponse.SerializeToString,
        ),
        "ExampleUnaryStream": slimrpc.unary_stream_rpc_method_handler(
            behaviour=servicer.ExampleUnaryStream,
            request_deserializer=pb2.ExampleRequest.FromString,
            response_serializer=pb2.ExampleResponse.SerializeToString,
        ),
        "ExampleStreamUnary": slimrpc.stream_unary_rpc_method_handler(
            behaviour=servicer.ExampleStreamUnary,
            request_deserializer=pb2.ExampleRequest.FromString,
            response_serializer=pb2.ExampleResponse.SerializeToString,
        ),
        "ExampleStreamStream": slimrpc.stream_stream_rpc_method_handler(
            behaviour=servicer.ExampleStreamStream,
            request_deserializer=pb2.ExampleRequest.FromString,
            response_serializer=pb2.ExampleResponse.SerializeToString,
        ),

    }

    server.register_method_handlers(
        "example_service.Test",
        rpc_method_handlers,
    )




# Copyright AGNTCY Contributors (https://github.com/agntcy)
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass

import slim_bindings


@dataclass
class SessionContext:
    """
    Context for RPC calls based on session information.
    """

    session_id: int
    source_name: str
    destination_name: str
    metadata: dict[str, str] | None = None

    @classmethod
    def from_session(cls, session: slim_bindings.Session) -> "SessionContext":
        """
        Create a SessionContext from session.
        """
        return cls(
            session_id=session.id,
            source_name=str(session.src),
            destination_name=str(session.dst),
            metadata=session.metadata,
        )


@dataclass
class MessageContext:
    """
    Context for RPC calls based on message information.
    """

    source_name: str
    destination_name: str
    payload_type: str
    metadata: dict[str, str] | None = None

    @classmethod
    def from_message_context(
        cls, msg_ctx: slim_bindings.MessageContext
    ) -> "MessageContext":
        """
        Create a MessageContext from message context.
        """
        return cls(
            source_name=str(msg_ctx.source_name),
            destination_name=str(msg_ctx.destination_name),
            payload_type=msg_ctx.payload_type,
            metadata=msg_ctx.metadata,
        )


# Keep Context as an alias for backward compatibility
Context = SessionContext



# Copyright AGNTCY Contributors (https://github.com/agntcy)
# SPDX-License-Identifier: Apache-2.0

import logging
from collections.abc import Callable
from dataclasses import dataclass
from typing import Iterable, Mapping, Optional, Union

from google.protobuf.any_pb2 import Any as pb_Any

logger = logging.getLogger(__name__)


class SRPCResponseError(Exception):
    def __init__(
        self,
        code: int,
        message: str,
        details: Optional[Iterable[Union[pb_Any, Mapping]]] = None,
    ) -> None:
        self.code = code
        self.message = message
        self.details = details
        super().__init__(message)


@dataclass
class RPCHandler:
    behaviour: Callable
    request_deserializer: Callable
    response_serializer: Callable
    request_streaming: bool = False
    response_streaming: bool = False


def unary_unary_rpc_method_handler(
    behaviour: Callable,
    request_deserializer: Callable = lambda x: x,
    response_serializer: Callable = lambda x: x,
) -> RPCHandler:
    return RPCHandler(
        behaviour=behaviour,
        request_deserializer=request_deserializer,
        response_serializer=response_serializer,
        request_streaming=False,
        response_streaming=False,
    )


def unary_stream_rpc_method_handler(
    behaviour: Callable,
    request_deserializer: Callable = lambda x: x,
    response_serializer: Callable = lambda x: x,
) -> RPCHandler:
    return RPCHandler(
        behaviour=behaviour,
        request_deserializer=request_deserializer,
        response_serializer=response_serializer,
        request_streaming=False,
        response_streaming=True,
    )


def stream_unary_rpc_method_handler(
    behaviour: Callable,
    request_deserializer: Callable = lambda x: x,
    response_serializer: Callable = lambda x: x,
) -> RPCHandler:
    return RPCHandler(
        behaviour=behaviour,
        request_deserializer=request_deserializer,
        response_serializer=response_serializer,
        request_streaming=True,
        response_streaming=False,
    )


def stream_stream_rpc_method_handler(
    behaviour: Callable,
    request_deserializer: Callable = lambda x: x,
    response_serializer: Callable = lambda x: x,
) -> RPCHandler:
    return RPCHandler(
        behaviour=behaviour,
        request_deserializer=request_deserializer,
        response_serializer=response_serializer,
        request_streaming=True,
        response_streaming=True,
    )



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class ExampleRequest(_message.Message):
    __slots__ = ("example_string", "example_integer")
    EXAMPLE_STRING_FIELD_NUMBER: _ClassVar[int]
    EXAMPLE_INTEGER_FIELD_NUMBER: _ClassVar[int]
    example_string: str
    example_integer: int
    def __init__(self, example_string: _Optional[str] = ..., example_integer: _Optional[int] = ...) -> None: ...

class ExampleResponse(_message.Message):
    __slots__ = ("example_string", "example_integer")
    EXAMPLE_STRING_FIELD_NUMBER: _ClassVar[int]
    EXAMPLE_INTEGER_FIELD_NUMBER: _ClassVar[int]
    example_string: str
    example_integer: int
    def __init__(self, example_string: _Optional[str] = ..., example_integer: _Optional[int] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: example.proto
# Protobuf Python Version: 5.29.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    3,
    '',
    'example.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rexample.proto\x12\x0f\x65xample_service\"`\n\x0e\x45xampleRequest\x12%\n\x0e\x65xample_string\x18\x01 \x01(\tR\rexampleString\x12\'\n\x0f\x65xample_integer\x18\x02 \x01(\x03R\x0e\x65xampleInteger\"a\n\x0f\x45xampleResponse\x12%\n\x0e\x65xample_string\x18\x01 \x01(\tR\rexampleString\x12\'\n\x0f\x65xample_integer\x18\x02 \x01(\x03R\x0e\x65xampleInteger2\xf2\x02\n\x04Test\x12V\n\x11\x45xampleUnaryUnary\x12\x1f.example_service.ExampleRequest\x1a .example_service.ExampleResponse\x12Y\n\x12\x45xampleUnaryStream\x12\x1f.example_service.ExampleRequest\x1a .example_service.ExampleResponse0\x01\x12Y\n\x12\x45xampleStreamUnary\x12\x1f.example_service.ExampleRequest\x1a .example_service.ExampleResponse(\x01\x12\\\n\x13\x45xampleStreamStream\x12\x1f.example_service.ExampleRequest\x1a .example_service.ExampleResponse(\x01\x30\x01\x42{\n\x13\x63om.example_serviceB\x0c\x45xampleProtoP\x01\xa2\x02\x03\x45XX\xaa\x02\x0e\x45xampleService\xca\x02\x0e\x45xampleService\xe2\x02\x1a\x45xampleService\\GPBMetadata\xea\x02\x0e\x45xampleServiceb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'example_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\023com.example_serviceB\014ExampleProtoP\001\242\002\003EXX\252\002\016ExampleService\312\002\016ExampleService\342\002\032ExampleService\\GPBMetadata\352\002\016ExampleService'
  _globals['_EXAMPLEREQUEST']._serialized_start=34
  _globals['_EXAMPLEREQUEST']._serialized_end=130
  _globals['_EXAMPLERESPONSE']._serialized_start=132
  _globals['_EXAMPLERESPONSE']._serialized_end=229
  _globals['_TEST']._serialized_start=232
  _globals['_TEST']._serialized_end=602
# @@protoc_insertion_point(module_scope)
